//
// Extract wav files from MDF/MDS files
// Copyright (c)2007 Bjorge Dijkstra (bjd@jooz.net)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// define this to one when compiling on big endian arch
#define ARCH_BIG_ENDIAN  0


#if ARCH_BIG_ENDIAN
#	define le_short(s) ( (((s) & 0xff) << 8)  | (((s) & 0xff00)  >> 8) )
#	define le_long(l)  ( (((s) & 0xff) << 24) | (((s) & 0xff000000) >> 24) | (((s) & 0xff00) << 8)  | (((s) & 0xff0000) >> 8) )
#else
#	define le_short(s) (s)
#	define le_long(l) (l)
#endif


//
// audio cd data
//   frames of 2352 bytes
//   each frame lasts 588/44100 seconds, i.e 75 frames/sec
//
#define CDDA_FRAME_SIZE 2352



#define MDS_DATABLOCK_SIZE	0x50
#define MDS_MAGIC "MEDIA DESCRIPTOR"

struct mds_header {
	char magic[16];	// string MEDIA DESCRIPTOR
	unsigned char major;
	unsigned char minor;
	short unk1;
	short num_sessions;
	short unk2;
	int   unk3[14];
	unsigned int   session_table_offset;
	int   unk4; // offset to some sort of table
};


struct mds_session {
	unsigned int   first_sector;
	unsigned int   last_sector;
	unsigned short session_num;
	unsigned char  num_data_blocks;
	unsigned char  num_leadin_blocks;
	unsigned short first_track;
	unsigned short last_track;
	unsigned int   unk4;
	unsigned int   data_blocks_offset;	
};


struct mds_track {
	unsigned short mode;
	unsigned short flags;
	unsigned int   track_num;
	unsigned char  time_hour;
	unsigned char  time_min;
	unsigned char  time_sec;
	unsigned char  time_frame;
	unsigned int   extra_data_offset;
	unsigned short sector_size;
	unsigned short unk1;
	unsigned int   unk2[4];
	unsigned int   start_sector;
	unsigned long long offset;
	unsigned int   unk3;
	unsigned int   file_idx_offset;
	int unk4[6];
};

struct mds_track_extra {
	unsigned int pregap;
	unsigned int length; // in sectors
};


/* wave file structs and constants */

#define RIFF_ID  0x46464952
#define WAVE_ID  0x45564157
#define FMT_ID   0x20746D66
#define DATA_ID  0x61746164

struct wav_header {
    int chunk_id;
    int chunk_size;
    int riff_type;
};

struct fmt_chunk {
    int chunk_id;
    int chunk_size;
    short compression;
    short channels;
    int rate;
    int bps;
    short align;
    short bits;
} __attribute__((packed))__;

struct data_chunk {
    int chunk_id;
    int chunk_size;
};

/**/



int arg_raw = 0;
int arg_info = 0;
int arg_file = 0;
int arg_track = 0;
int arg_quiet = 0;
int arg_dump = 0;
int arg_stdout = 0;
char arg_filename[1024];




void hexdump(void * buffer, int length) {
    int i = 0;
    int j;
    
    while (i < length) {
	printf("%04X   ", i);
	for (j=0; j<16 && i<length; j++) {
	    printf("%02X ", ((unsigned char*)buffer)[i]);
	    i++;
	    if ((i % 4) == 0) {
		printf(" ");
	    }
	}
	printf("\n");
    }
}



#define EXTRACT_BUFFER_SIZE  (16*CDDA_FRAME_SIZE)

int mdf_extract(char * mdffile, int num, int start, int length)
{
    int dup_name = 0;
    char *buffer = 0;
    int len;
    FILE * fin = NULL, *fout = NULL;
    char outfilename[1024];
    
    if (strchr(mdffile, '*') != NULL) {
	// no explicit filename indicated, assume same as .mds file but
	// with .mdf extension instead
	mdffile = strdup(arg_filename);
	mdffile[strlen(mdffile)-1] = 'f';
	dup_name = 1;
    }
    
    if (arg_raw)
	sprintf(outfilename, "track_%02d.raw", num);
    else
	sprintf(outfilename, "track_%02d.wav", num);
	
    printf("extracting track %d from %s to %s (%d bytes)\n", num, mdffile, outfilename, length * CDDA_FRAME_SIZE);
    
    fin = fopen(mdffile, "r");
    if (arg_stdout)
	fout = stdout;
    else
	fout = fopen(outfilename, "w");
    
    if (fin && fout) {
	buffer = malloc(EXTRACT_BUFFER_SIZE);
	start  = start * CDDA_FRAME_SIZE;
	length = length * CDDA_FRAME_SIZE;	
	
	if (fseek(fin, start, SEEK_SET) < 0) {
	    perror("fseek");
	}
	
	if (arg_raw == 0) {
	    struct wav_header wh;
	    struct fmt_chunk fc;
	    struct data_chunk dc;
	    	    
	    /* write wav header */
	    wh.chunk_id = le_long(RIFF_ID);
	    wh.chunk_size = le_long(length + 16 + 8);
	    wh.riff_type = le_long(WAVE_ID);
	    
	    /* format chunk */
	    fc.chunk_id = le_long(FMT_ID);
	    fc.chunk_size = le_long(16);
	    fc.compression = le_short(1);
	    fc.channels = le_short(2);
	    fc.rate = le_long(44100);
	    fc.bps = le_long(44100 * 2 * 2);
	    fc.align = le_short(4);
	    fc.bits = le_short(16);
	    
	    /* data chunk */
	    dc.chunk_id = le_long(DATA_ID);
	    dc.chunk_size = le_long(length);
	    
	    fwrite(&wh, 1, sizeof(wh), fout);
	    fwrite(&fc, 1, sizeof(fc), fout);
	    fwrite(&dc, 1, sizeof(dc), fout);    
	}
	
	while (length > 0) {
	    len = (length > EXTRACT_BUFFER_SIZE) ? EXTRACT_BUFFER_SIZE : length;
	    len = fread(buffer, 1, len, fin);
	    if (len >= 0) {
		if (fwrite(buffer, 1, len, fout) >= 0) {
		    length = length - len;
		} else {
		    perror("fwrite");
		    length = 0;
		}
	    } else {
		perror("fread");
		length = 0;
	    }
	}
    }
    
    if (fin)
	fclose(fin);
	
    if (fout && !arg_stdout)
	fclose(fout);
    
    if (dup_name)
	free(mdffile);
	
    if (buffer)
	free(buffer);
	
    return 0;
}


void mds_walk_track(void * buffer, struct mds_track *t)
{
	struct mds_track_extra *e;
	int length;
	

	if (t->track_num < 0xa0) {
		if (t->extra_data_offset) {
			e = buffer + t->extra_data_offset;
			length =  e->length;
		} else
			length = 0;
			
		if (arg_track == 0 || arg_track == t->track_num) {
		
		    if (arg_dump) {
			hexdump(t, sizeof(*t));
		    }
		
		    if (arg_info) {
			printf("\t\tTrack: %2d, Position: %02d:%02d:%02d.%02d, sector: %d, length: %d\n",
			    t->track_num,
			    t->time_hour, t->time_min,
			    t->time_sec, t->time_frame,
			    t->start_sector, e->length
			);
		    } else {		    
			mdf_extract(buffer + t->file_idx_offset + 16, t->track_num, t->start_sector, length);
		    }				    
		}
	}
}


void mds_walk_session(void * buffer, struct mds_session *s)
{
	int i;
	struct mds_track *t;

	if (arg_dump) {
	    hexdump(s, sizeof(*s));
	}
	
	if (arg_info) {
	    printf("\tsession: %d\n", s->session_num);
	    printf("\t    first sector: %d\n", s->first_sector);
	    printf("\t    last sector : %d\n", s->last_sector);
    	    printf("\t    blocks: %d, leadin: %d, first track: %d, last track: %d\n",
		s->num_data_blocks, s->num_leadin_blocks,
		s->first_track, s->last_track);
	}
	
	t = (void*)buffer + s->data_blocks_offset;
	for (i = 0; i < s->num_data_blocks; i++) {
		mds_walk_track(buffer, &t[i]);
	}	
}


void mds_walk(void * buffer, struct mds_header * h)
{
	int i;
	struct mds_session *s;

	if (arg_info)
	    printf("MDS version: %d.%d\n", h->major, h->minor);

	s = (void*)buffer + h->session_table_offset;

	for (i = 0; i < h->num_sessions; i++) {
		mds_walk_session(h, &s[i]);
	}
}


void usage(void)
{
    printf("mdfextract 0.1	(c)2007 Bjorge Dijkstra\n\n");
    printf("usage: mdfextract [options] file.mds\n\n");
    printf("options:\n");
    printf("-i\t\toutput mds info instead of extracting\n");
    printf("-r\t\textract raw track data\n");
    printf("-s\t\toutput to stdout instead of file\n");
    printf("-t <num>\textract single track (num > 0)\n");
    printf("-q\t\tquiet, no output\n");
}


int parse_cmdline(int argc, char * argv[])
{
    int i = 1;
    int err = 0;

    if (argc <= 1) {
	err = 1;
    }
        
    while (!err && i < argc) {
	if (strcmp(argv[i], "-i") == 0) {
	    arg_info = 1;
	} else if (strcmp(argv[i], "-r") == 0) {
	    arg_raw = 1;
	} else if (strcmp(argv[i], "-q") == 0) {
	    arg_quiet = 1;
	} else if (strcmp(argv[i], "-s") == 0) {
	    arg_stdout = 1;	    
	} else if (strcmp(argv[i], "-t") == 0) {
	    i++;
	    if (i < argc) {
		if (sscanf(argv[i], "%d", &arg_track) != 1) 
		    err = 1;
		 else 
		    if (arg_track == 0) 
			err = 1;
	    } else 
		err = 1;
	} else if (strcmp(argv[i], "-d") == 0) {
	    arg_dump = 1;
	} else if (!arg_file) {
	    arg_file = 1;
	    strncpy(arg_filename, argv[i], sizeof(arg_filename));
	} else {
	    /* extra data on cmdline */
	    err = 1;
	}
	i++;
    }
    if (!arg_file) {
	/* missing filename */
	err = 1;
    }
    
    return err;
}




int main(int argc, char *argv[])
{
	char * buffer = 0;
	FILE * f;

	if (parse_cmdline(argc, argv) != 0) {	
	    usage();
	    exit(0);
	}
	
	f = fopen(arg_filename, "r");
	if (f) {
	    buffer = malloc(64 * 1024);
	    fread(buffer, 1, 64 * 1024, f);
	    fclose(f);
	    if (strncmp(buffer, MDS_MAGIC, 16) == 0)
    		mds_walk(buffer, (struct mds_header *) buffer);			
	}

	if (buffer)
	    free(buffer);
	return 0;
}

